import numpy as np
from .constants import *
from scipy.interpolate import interp1d
from tqdm import tqdm
import pickle

##########
# 

##########

class Event:
    '''
    The Event class contains all information about a single particle event in the detector. 
    This starts with information of energy depositions imported from Geant4 (pos,dE,times). 
    Electron and hole drift paths are stored and used to determine the signal on a detector. 

    Currently this only works for a single contact, but we plan to extend to all contacts.
    '''
    def __init__(self, _id, _pos, _dE, _times):
        self.ID = _id
        self.pos = _pos
        self.dE = _dE
        self.times = _times

        self.dQ = []
        self.dI = []
        self.dt = []

        self.quasiparticles = []
        
        self.signal_I = None
        self.signal_times = None
        
    def convolveElectronicResponse(self, electronicResponse):
        func_I = interp1d(self.dt, self.dI, bounds_error=False, fill_value=0)
        temp_times = electronicResponse["times"]
        dt = np.diff(temp_times)[0]
        
        self.signal_I = np.convolve(func_I(temp_times),electronicResponse["step"])
        self.signal_times = np.arange(0,len(self.signal_I)*dt, dt)
        
        return None
        
    def addGaussianNoise(self, sigma=1, SNR=None):
        if SNR is not None:
            sigma = np.max(np.abs(self.signal_I))/SNR
        self.signal_I += np.random.normal(0, sigma , len(self.signal_I))   
        
    def convertUnits(self, energyConversionFactor, lengthConversionFactor, timeConversionFactor):
        self.dE *= energyConversionFactor
        self.pos *= lengthConversionFactor
        self.times *= timeConversionFactor
        return None
        
    def calculateInducedCurrent(self, weightingField, dt, interp3d=True):
        weightingFieldx_interp, weightingFieldy_interp, weightingFieldz_interp, weightingFieldMag_interp = weightingField.interpolate(interp3d=interp3d)

        max_time = max([o.time[-1] for o in self.quasiparticles])
        
        times_I = np.arange(0,max_time,dt)
        
        induced_I = np.zeros(len(times_I))

        for i in range(len(self.quasiparticles)):
            o = self.quasiparticles[i]
            Is = [o.q*np.dot(o.vel[j],
                        [weightingFieldx_interp(o.pos[j]),
                         weightingFieldy_interp(o.pos[j]),
                         weightingFieldz_interp(o.pos[j])]) for j in range(len(o.pos))]

            if len(Is) > 0:
                func_I = interp1d(o.time, Is, bounds_error=False, fill_value=0)
                induced_I += func_I(times_I)

        '''for i in tqdm(range(len(self.vel_drift_e))):
            if interp3d:
                Is = [-qe_SI*np.dot(self.vel_drift_e[i][j], 
                        [weightingFieldx_interp(self.pos_drift_e[i][j]),
                         weightingFieldy_interp(self.pos_drift_e[i][j]),
                         weightingFieldz_interp(self.pos_drift_e[i][j])]) for j in range(len(self.vel_drift_e[i]))]
                         
            else:
                Is = [-qe_SI*np.dot(self.vel_drift_e[i][j], 
                        [weightingFieldx_interp(self.pos_drift_e[i][j]),
                         weightingFieldy_interp(self.pos_drift_e[i][j]),
                         weightingFieldz_interp(self.pos_drift_e[i][j])])[0] for j in range(len(self.vel_drift_e[i]))]

            times = self.times_drift_e[i][:-1]

            if len(Is)>1:
                func_I = interp1d(times, Is, bounds_error=False, fill_value=0)
                induced_I += func_I(times_I)
                
        for i in tqdm(range(len(self.vel_drift_h))):
            if interp3d:
                Is = [qe_SI*np.dot(self.vel_drift_h[i][j], 
                        [weightingFieldx_interp(self.pos_drift_h[i][j]),
                         weightingFieldy_interp(self.pos_drift_h[i][j]),
                         weightingFieldz_interp(self.pos_drift_h[i][j])]) for j in range(len(self.vel_drift_h[i]))]
                         
            else:
                Is = [qe_SI*np.dot(self.vel_drift_h[i][j], 
                        [weightingFieldx_interp(self.pos_drift_h[i][j]),
                         weightingFieldy_interp(self.pos_drift_h[i][j]),
                         weightingFieldz_interp(self.pos_drift_h[i][j])])[0] for j in range(len(self.vel_drift_h[i]))]
            times = self.times_drift_h[i][:-1]

            if len(Is)>1:
                func_I = interp1d(times, Is, bounds_error=False, fill_value=0)
                induced_I += func_I(times_I)'''
        
        self.dI = induced_I
        self.dt = times_I
        
        return None    
    
    def sample(self, dt,length=None):
        '''
        Samples the induced current of an event to larger timesteps so that it is the same format as nab data. 
        There is some question as to the proper way to do this; for now we simply take the value at the exact time sampled.
        '''
        step = int(dt/(self.signal_times[1]-self.signal_times[0])) 
        signal = self.signal_I[::step].copy()
        if length==None:
            return signal
        else:
            return np.pad(signal, (round(length/2),round(length/2)-len(signal)),"edge")

def eventsFromG4root(filename):
    import uproot
    import pandas

    file = uproot.open(filename)
    tree = file["ntuple/hits"]

    try:
        df = tree.arrays(["eventID", "trackID", "x", "y", "z", "time", "eDep"], 
                         library="pd")
        gdf = df.groupby("eventID")
    except:
        #old formatting exception
        df = tree.arrays(["iD", "x", "y", "z", "time", "eDep"], library="pd")
        gdf = df.groupby("iD")

    events = []
    for eID, group in gdf:
        event = Event(eID, group[["x", "y", "z"]].to_numpy(), group["eDep"].to_numpy(), group["time"].to_numpy())
        event.convertUnits(1e3,1e-3,1e-9)
        events.append(event)
    return events
    
#convert events to nabPy form, and save optionally save pickle file if filename is given. dt is time sampling in ns, which is 4ns for current nab daq
def saveEventsNabPy(events, filename=None, dt=4e-9, length=7000):
    new_events =np.array([event.sample(dt,length) for event in events])
    if filename is not None:
        with open(filename+".pkl", "wb") as file:
            pickle.dump(new_events, file)
    
    return new_events
    
def loadEventsNabPy(filename):
    with open(filename, "rb") as file:
        events = pickle.load(file)
    
    return events
    
    
#save complete event object, for example if you want to be able to re-downsample, apply a different convolution, etc.
def saveEvents(events, filename):
    with open(filename+".pkl", "wb") as file:
        pickle.dump(events, file)
    return
    
def loadEvents(filename):
    with open(filename+".pkl", "rb") as file:
        events = pickle.load(file)
    return events
    



from .constants import *
from numba import jit
import numpy as np
from math import *

#Note everything is in meters. This is not standard in the literature on mobility.

beta_el = 1 # beta coefficient in electron mobility parametrization according to Knoll
beta_h = 2
vstar_el = 2.4e5 # m/s
Theta_el = 600 # K
C_el = 0.8
# TODO: Hole values are just sort of copy pasted, needs to be actually checked
vstar_h = 1e5 # m/s
Theta_h = 600 # K
C_h = 0.8
D0_el = 0.0036 # m^2/s
D0_h = 0.0012 # m^2/s

@jit(nopython=True)
def mu0_el(T):
    return 0.1440 * (T/300)**-2.01 # m^2/V/s

@jit(nopython=True)
def mu0_h(T):
    return 0.0450 * (T/300)**-2.01 # m^2/V/s

# Temperature-dependent saturation velocity
@jit(nopython=True)
def vs_el(T): 
    return vstar_el/(1+C_el*exp(T/Theta_el))

@jit(nopython=True)
def vs_h(T):
    return vstar_h/(1+C_h*exp(T/Theta_h))

@jit(nopython=True)
def diffusion_electron(T):
    return D0_el*(T/300)**-1.01

@jit(nopython=True)
def diffusion_hole(T):
    return D0_h*(T/300)**-1.01

def mobility(mu, E, beta, vs):
    return mu/(1+(mu*E/vs)**(1/beta))**beta

# Impurity density dependent mobility parametrization according to Klaassen
mu_max = mu0_el(300)
mu_min = 0.00685 # m^2/V/s
Nref1 = 9.2e10 # m^-3
Nref2 = 3.41e14 # m^-3
alpha1 = 0.711

@jit(nopython=True, fastmath=True)
def mu_I_N(T): 
    return mu_max**2/(mu_max-mu_min)*(T/300)**(3*alpha1-1.5)

@jit(nopython=True, fastmath=True)
def mu_I_c(T):
    return mu_min*mu_max/(mu_max-mu_min)*(300/T)**0.5

@jit(nopython=True, fastmath=True)
def mu_I(T, NI):
    return mu_I_N(T)*(Nref1/NI)**alpha1 + mu_I_c(T)

def generalized_mobility_el(T, NI, E):
    mu = 1/(1/mu0_el(T)+1/mu_I(T, NI))
    vsat = vs_el(T)
    return mu/(1+(mu*E/vsat)**(1/beta_el))**beta_el

@jit(nopython=True)
def generalized_mobility_h(T, NI, E):
    mu = 1/(1/mu0_h(T)+1/mu_I(T, NI))
    vsat = vs_h(T)
    return mu/(1+(mu*E/vsat)**(1/beta_h))**beta_h

def generalized_diffusion_el(T, NI, E):
    return mobility(T, NI, E)*T*kB



from .constants import *
from .mobility import *
import numpy as np

def CoulombForce(q1, q2, pos1, pos2, safety = 7e-9):
    r2 = max(safety**2, sum((pos1-pos2)**2))
    Fs = -kE_SI*q1*q2/r2/epsR_Si
    return Fs*(pos2-pos1)/r2**0.5

def getEffectiveCoulombField(objects):
    Ecoul = np.zeros((len(objects), 3))
    for i in range(len(objects)):
        Q = objects[i].q
        F = np.zeros(3)
        for j in range(len(objects)):
            if i != j:
                F+= CoulombForce(Q, objects[j].q, objects[i].pos[-1], objects[j].pos[-1])
            Ecoul[i] = F/Q
    return Ecoul

def insideBoundaryCheck(pos, bounds):
    return ((pos[0] >= bounds[0][0]) & (pos[0] <= bounds[0][1]) & (pos[1] >= bounds[1][0]) & (pos[1] <= bounds[1][1])
            & (pos[2] >= bounds[2][0]) & (pos[2] <= bounds[2][1]))

def updateQuasiParticles(objects, eps, maxdt, Ex_i, Ey_i, Ez_i, E_i, bounds, temp, diffusion=True, coulomb=False, tauTrap = lambda x, y, z : 1e9, NI = lambda x, y, z : 1e10):
    #Vector to save the effective electric field for each particle at the time step
    Eeff = np.zeros((len(objects), 3))

    # If coulomb is enabled, an effective electric field because of all other charges is added (plasma effects)
    if coulomb:
        Eeff += getEffectiveCoulombField(objects)

    for i in range(len(objects)):
        pos = objects[i].pos[-1]
        dt = maxdt

        Eeff[i] += np.array([Ex_i(pos), Ey_i(pos), Ez_i(pos)])

        if objects[i].q < 0:
            mu = generalized_mobility_el(temp, NI(*pos), np.linalg.norm(Eeff[i]))
            dv = -mu*Eeff[i]
        else:
            mu = generalized_mobility_h(temp, NI(*pos), np.linalg.norm(Eeff[i]))
            dv = mu*Eeff[i]

        if not coulomb:
            dt = min(abs(eps/np.linalg.norm(dv)), maxdt)

        if diffusion:
            dv += ((diffusion_electron(temp) if objects[i].q < 0 else diffusion_hole(temp))/dt)**0.5*np.random.normal(size=3)

        objects[i].addTime(objects[i].time[-1]+dt)
        objects[i].addVel(dv)
        objects[i].addPos(objects[i].pos[-1]+objects[i].vel[-1]*dt)

        #Check if particles are still inside the specified boundaries, otherwise kill them
        objects[i].alive = insideBoundaryCheck(objects[i].pos[-1], bounds)

        #Check whether a particle is captured. If so, the particle is killed. 
        #dt must be substantially smaller than tauTrap in order for Poisson process to work
        p = dt/tauTrap(*(objects[i].pos[-1]))
        r = np.random.random()
        if r < p:
            objects[i].alive = False

    return objects


'''def propagateCarrier(x0, y0, z0, eps, Ex_i, Ey_i, Ez_i, E_i, bounds, T, tauTrap = lambda x,y,z: 1e9, d=None, NI=lambda x, y,z: 1e10, diffusion=False, electron=True, interp3d=False):
    rand_steps = np.random.standard_normal(size=(10000,3))
    x = [x0, ]
    y = [y0, ]
    z = [z0, ]
    t = [0, ]
    diffRatio = [0, ]
    while ((x[-1] >= bounds[0][0]) & (x[-1] <= bounds[0][1]) & (y[-1] >= bounds[1][0]) & (y[-1] <= bounds[1][1]) 
            & (z[-1] >= bounds[2][0]) & (z[-1] <= bounds[2][1])):
        #print(len(t), len(t)<stepLimit)
        #print(x[-1], y[-1], z[-1])
        
        if interp3d:
            E = E_i([x[-1], y[-1], z[-1]])
            Ex = Ex_i([x[-1], y[-1], z[-1]])
            Ey = Ey_i([x[-1], y[-1], z[-1]])
            Ez = Ez_i([x[-1], y[-1], z[-1]])
        
        else:
            E = E_i([x[-1], y[-1], z[-1]])[0]
            Ex = Ex_i([x[-1], y[-1], z[-1]])[0]
            Ey = Ey_i([x[-1], y[-1], z[-1]])[0]
            Ez = Ez_i([x[-1], y[-1], z[-1]])[0]
        
        
        if electron:
        	mu = generalized_mobility_el(T, NI(x[-1], y[-1], z[-1]), E)
        else:
        	mu = mu0_h(T)
        	
        if d == None:
            #d_temp = D(T, NI(x[-1], y[-1], z[-1]), E)
            d_temp = diffusion_electron(T) if electron else diffusion_hole(T)
        else:
            d_temp = d


        dx = 0
        dy = 0
        dz = 0
        
        if abs(E) > 0:
            dt = min(min(abs(eps/mu/E), tauTrap(x[-1], y[-1], z[-1])/1e5),(bounds[2][1]-z[-1])/mu/np.abs(Ez)+1e-10) #Time step needs to be significantly smaller than trapping time for Poisson process to make sense 
            
            sign = -1 if electron else 1
            dx = sign*dt*mu*Ex
            dy = sign*dt*mu*Ey
            dz = sign*dt*mu*Ez
            dr = np.sqrt(dx**2+dy**2+dz**2)
            
            #print(dt, dz)
            angle = 0
            if (dx != 0):
                angle = np.arctan(dy/dx)
            phi = np.arccos(dz/dr)
        else:
            dt = tauTrap(x[-1], y[-1], z[-1])/1e5
            angle = 0
        
        #Check if particle gets trapped
        #p = dt/tauTrap(x[-1], y[-1], z[-1])
        #r = np.random.random()
        #if r < p:
        #    return x, y, z, t
        
        
        dxd = 0
        dyd = 0
        dzd = 0
       
        if (diffusion and d_temp != 0):
                dxd, dyd, dzd = getDriftStep(angle, d_temp, dt)
                #print(rand_steps[0])
                #dxd, dyd, dzd = getDriftStep2(angle, d_temp, dt, rand_steps[0])
                #rand_steps = np.roll(rand_steps, 1, axis=0)
        
        #print("diffusion", dxd, dyd, dzd, dt)
        
        x.append(x[-1] + dx+dxd)
        y.append(y[-1] + dy+dyd)
        z.append(z[-1] + dz+dzd)
        t.append(t[-1] + dt)
        
        #print(x[-1], y[-1], t[-1])
        
    return x, y, z, t'''

if __name__ == "__main__":
    x, y, z, t = propagateCarrier(0, 0, 0, 1e-4, lambda a : [0,], lambda a : [0,], lambda a: [-750e2,], lambda a : [-750e2,], [[-1, 1], [-1, 1], [0, 0.002]], 130, diffusion=True)
